/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/

#include <proxy/rpc/forward_metadata.hpp>
#include <global/rpc/distributor.hpp>
#include <global/rpc/rpc_types.hpp>

using namespace std;

namespace gkfs {
namespace rpc {

int forward_create(const std::string& path, const mode_t mode) {
    hg_handle_t rpc_handle;
    rpc_mk_node_in_t daemon_in{};
    rpc_err_out_t daemon_out{};
    int err;
    // fill in
    daemon_in.path = path.c_str();
    daemon_in.mode = mode;
    // Create handle
    PROXY_DATA->log()->debug("{}() Creating Margo handle ...", __func__);
    auto endp = PROXY_DATA->rpc_endpoints().at(PROXY_DATA->distributor()->locate_file_metadata(path));
    auto ret = margo_create(PROXY_DATA->client_rpc_mid(), endp,
                            PROXY_DATA->rpc_client_ids().rpc_create_id, &rpc_handle);
    if (ret != HG_SUCCESS)
        return EBUSY;
    ret = margo_forward(rpc_handle, &daemon_in);
    if (ret == HG_SUCCESS) {
        // Get response
        PROXY_DATA->log()->trace("{}() Waiting for response", __func__);
        ret = margo_get_output(rpc_handle, &daemon_out);
        if (ret == HG_SUCCESS) {
            PROXY_DATA->log()->debug("{}() Got response success: {}", __func__, daemon_out.err);
            err = daemon_out.err;
            margo_free_output(rpc_handle, &daemon_out);
        } else {
            // something is wrong
            err = EBUSY;
            PROXY_DATA->log()->error("{}() while getting rpc output", __func__);
        }
    } else {
        // something is wrong
        err = EBUSY;
        PROXY_DATA->log()->error("{}() sending rpc", __func__);
    }

    /* clean up resources consumed by this rpc */
    margo_destroy(rpc_handle);
    return err;
}

std::pair<int, std::string> forward_stat(const std::string& path) {
    hg_handle_t rpc_handle;
    rpc_path_only_in_t daemon_in{};
    rpc_stat_out_t daemon_out{};
    int err;
    string attr{};
    // fill in
    daemon_in.path = path.c_str();
    // Create handle
    PROXY_DATA->log()->debug("{}() Creating Margo handle ...", __func__);
    auto endp = PROXY_DATA->rpc_endpoints().at(PROXY_DATA->distributor()->locate_file_metadata(path));
    auto ret = margo_create(PROXY_DATA->client_rpc_mid(), endp,
                            PROXY_DATA->rpc_client_ids().rpc_stat_id, &rpc_handle);
    if (ret != HG_SUCCESS)
        return make_pair(EBUSY, attr);
    ret = margo_forward(rpc_handle, &daemon_in);
    if (ret == HG_SUCCESS) {
        // Get response
        PROXY_DATA->log()->trace("{}() Waiting for response", __func__);
        ret = margo_get_output(rpc_handle, &daemon_out);
        if (ret == HG_SUCCESS) {
            PROXY_DATA->log()->debug("{}() Got response success: {}", __func__, daemon_out.err);
            err = daemon_out.err;
            if (err == 0)
                attr = daemon_out.db_val;
            margo_free_output(rpc_handle, &daemon_out);
        } else {
            // something is wrong
            err = EBUSY;
            PROXY_DATA->log()->error("{}() while getting rpc output", __func__);
        }
    } else {
        // something is wrong
        err = EBUSY;
        PROXY_DATA->log()->error("{}() sending rpc", __func__);
    }

    /* clean up resources consumed by this rpc */
    margo_destroy(rpc_handle);
    return make_pair(err, attr);
}

} // namespace rpc
} // namespace gkfs
